@@ -24,6 +24,8 @@ class Settings:
   """ If set to true, reactions such as A + B -> A are allowed"""
   addDegradationSteps = False;
   """Set true if you want every floating node (not boundary nodes) to have a degradation step"""
+  removeBoundarySpecies = True
+  """Set true if you want and sink and source species to be classed as boundary species"""
 
   @dataclass
   class ReactionProbabilities:
@@ -110,8 +112,9 @@ def getLinearChain (lengthOfChain, rateLawType='MassAction', keqRatio=5):
     # Initialize values
     if rateLawType == 'MassAction':
        for i in range (n):
-           model += 'k' + str (i+1) + '0 = ' + str ('{:.2f}'.format (_random.random()*keqRatio)) + ';  ' + \
-           'k' + str (i+1) + '1 = ' + str ('{:.2f}'.format (_random.random()*1)) + '\n'
+           # Add 0.01 to ensure the value won't be zero
+           model += 'k' + str (i+1) + '0 = ' + str ('{:.2f}'.format ((_random.random()+0.01)*keqRatio)) + ';  ' + \
+           'k' + str (i+1) + '1 = ' + str ('{:.2f}'.format ((_random.random()+0.01)*1)) + '\n'
            
     model += 'Xo = ' + str ('{:.2f}'.format (_random.randint(1, 10))) + '\n' 
     model += 'X1 = 0' + '\n' 
@@ -187,6 +190,7 @@ def _pickReactionType():
 # Generates a reaction network in the form of a reaction list
 # reactionList = [numSpecies, reaction, reaction, ....]
 # reaction = [reactionType, [list of reactants], [list of products], rateConstant]
+# Doesn't differentiate between boundary and floating species
 # Disallowed reactions:
 # S1 -> S1
 # S1 + S2 -> S2  # Can't have the same reactant and product unless allowMassViolatingReactions is true
@@ -219,6 +223,8 @@ def _generateReactionList (nSpecies, nReactions):
              species = range (nSpecies)
              # Remove reactant1 and 2 from the species list
              species = _np.delete (species, [reactant1, reactant2], axis=0)
+             if len (species) == 0:
+                raise Exception("Unable to pick a species why mainting mass conservation")
              # Then pick a product from the reactants that are left
              product = species[_random.randint (0, len (species)-1)]
                
@@ -235,6 +241,9 @@ def _generateReactionList (nSpecies, nReactions):
              species = range (nSpecies)
              # Remove reactant1 from the species list
              species = _np.delete (species, [reactant1], axis=0)
+             if len (species) == 0:
+                raise Exception("Unable to pick a species why mainting mass conservation")
+
              # Then pick a product from the reactants that are left
              product1 = species[_random.randint (0, len (species)-1)]
              product2 = species[_random.randint (0, len (species)-1)]
@@ -253,6 +262,8 @@ def _generateReactionList (nSpecies, nReactions):
              species = range (nSpecies)
              # Remove reactant1 and 2 from the species list
              species = _np.delete (species, [reactant1, reactant2], axis=0)
+             if len (species) == 0:
+                raise Exception("Unable to pick a species why mainting mass conservation")             
              # Then pick a product from the reactants that are left
              product1 = species[_random.randint (0, len (species)-1)]
              product2 = species[_random.randint (0, len (species)-1)]
@@ -260,18 +271,25 @@ def _generateReactionList (nSpecies, nReactions):
           element = [rt, [reactant1, reactant2], [product1, product2], rateConstant]
           reactionList.append (element)            
 
-    reactionList.insert (0, nReactions)
+    reactionList.insert (0, nSpecies) 
     return reactionList
     
 
 # Includes boundary and floating species
 # Returns a list:
 # [New Stoichiometry matrix, list of floatingIds, list of boundaryIds]
+# On entry, reactionList has the structure (obtained by calling _generateReactionList)
+# reactionList = [numSpecies, reaction, reaction, ....]
+# reaction = [reactionType, [list of reactants], [list of products], rateConstant]
+
 def _getFullStoichiometryMatrix (reactionList):
     
     nSpecies = reactionList[0]
     reactionListCopy = _copy.deepcopy (reactionList)
+    # Remove the first entry in the list which is the number of species
+    # This just makes it easier to index
     reactionListCopy.pop (0)
+    # Prepare space for the stoichiometry matrix
     st = _np.zeros ((nSpecies, len(reactionListCopy)))
     
     for index, r in enumerate (reactionListCopy):
@@ -285,31 +303,31 @@ def _getFullStoichiometryMatrix (reactionList):
         if r[0] == TReactionType.BIUNI:
             # BiUni
             reactant1 = reactionListCopy[index][1][0]
-            st[reactant1, index] = -1
+            st[reactant1, index] += -1
             reactant2 = reactionListCopy[index][1][1]
-            st[reactant2, index] = -1
+            st[reactant2, index] += -1
             product = reactionListCopy[index][2][0]
             st[product, index] = 1
 
         if r[0] == TReactionType.UNIBI:
             # UniBi
             reactant1 = reactionListCopy[index][1][0]
-            st[reactant1, index] = -1
+            st[reactant1, index] += -1
             product1 = reactionListCopy[index][2][0]
-            st[product1, index] = 1
+            st[product1, index] += 1
             product2 = reactionListCopy[index][2][1]
-            st[product2, index] = 1
+            st[product2, index] += 1
 
         if r[0] == TReactionType.BIBI:
             # BiBi
             reactant1 = reactionListCopy[index][1][0]
-            st[reactant1, index] = -1
+            st[reactant1, index] += -1
             reactant2 = reactionListCopy[index][1][1]
-            st[reactant2, index] = -1
+            st[reactant2, index] += -1
             product1 = reactionListCopy[index][2][0]
-            st[product1, index] = 1
+            st[product1, index] += 1
             product2 = reactionListCopy[index][2][1]
-            st[product2, index] = 1
+            st[product2, index] += 1
 
     return st   
         
@@ -402,7 +420,7 @@ def _getAntimonyScript (floatingIds, boundaryIds, reactionList, isReversible):
            antStr = antStr + 'S' + str (reactionListCopy[reactionIndex][2][0])
            antStr = antStr + ' + '
            antStr = antStr + 'S' + str (reactionListCopy[reactionIndex][2][1])
-           antStr = antStr + '; E' + str (index) + '*(k' + str (reactionIndex) + '*S' + str (reactionListCopy[reactionIndex][1][0])
+           antStr = antStr + '; E' + str (reactionIndex) + '*(k' + str (reactionIndex) + '*S' + str (reactionListCopy[reactionIndex][1][0])
            if isReversible:
              antStr = antStr + ' - k' + str (reactionIndex) + 'r' + '*S' + str (reactionListCopy[reactionIndex][2][0]) + '*S' + str (reactionListCopy[reactionIndex][2][1])
            antStr = antStr + ')'
@@ -486,7 +504,13 @@ def getRandomNetworkDataStructure (nSpecies, nReactions, isReversible=False, ran
     rl = _generateReactionList (nSpecies, nReactions)  
     st = _getFullStoichiometryMatrix (rl)
  
-    stt = _removeBoundaryNodes (st)
+    if Settings.removeBoundarySpecies:
+       # Note: stt = [stoich Matrix, floatIds, boundaryIds] 
+       stt = _removeBoundaryNodes (st)
+       if returnStoichiometryMatrix:
+          return stt[0]
+    else:
+       stt = [[], _np.arange (nSpecies), []]     
      
     if Settings.addDegradationSteps:
        for sp in stt[1]:
@@ -544,9 +568,13 @@ def getRandomNetwork (nSpecies, nReactions, isReversible=False, returnStoichiome
     if returnFullStoichiometryMatrix:
        return st
     
-    stt = _removeBoundaryNodes (st)
-    if returnStoichiometryMatrix:
-       return stt[0]
+    if Settings.removeBoundarySpecies:
+       # Note: stt = [stoich Matrix, floatIds, BoundaryIds] 
+       stt = _removeBoundaryNodes (st)
+       if returnStoichiometryMatrix:
+          return stt[0]
+    else:
+       stt = [[], _np.arange (nSpecies), []] 
       
     # stt[1] = floating species Ids
     # stt[2] = boundary species Ids    
